#include <string>
#include <iostream>
#include "graph.h"
using namespace std;



// ********************************************* PATH ***************************************************************************

//constructor
Path::Path(const int A[], int k)
{
    if(k>0) //positive k - for version 1: we are copying verticex A[1] to P[1], A[2] to P[2], etc.
    {
      size=k;
      for (int i=1;i<=k;i++)
        P[i]=A[i];
    }
    else{  //negative k - for version 2: we are copying vertex  A[0] to P[1], A[1] to P[2], etc.
      size=-k;
      for(int i=0; i<size; i++)
        P[i+1]=A[i];
    }

//calculate hash value: binary number with j-th bit set to 1 <=> path contains vertex j
    hash=0;
    for(int i=1;i<=size;i++)
        hash|= (1<<(P[i]-1));
}

//for testing: displaying the path
void Path::showPath()
{
  cout<<"<"<<size<<">";
  for (int i=1;i<=size-1;i++)
    cout<<P[i]<<",";
    cout<<P[size]<<" ";
}


//checking whether a path P has a common vertex with the setSize set stored in the table C
//the simplest algorithm
// complexity: |P|*|C| <= |V|/2*|V|/2 = |V|^2 / 4, where |V| is a number of vertices in the graph G
inline bool Path::hasCommonVertex (int C[], int setSize)
{
  bool wynik=false;
  int i,j;
  for (i=1;i<=size && !wynik;i++)
    for (j=1;j<=setSize;j++)
      wynik=wynik || (C[j]==P[i]);
  return wynik;

}

// ********************************** GRAPH *************************************************************************************

// Implementation of the constructor
Graph::Graph(string s, int size)
{
    char c;		         // a help variable
    int i=1, j=1;        // actual indices of A
    int length;	         // the length of s
    
    
    this->graphSize=size;
    
    length=s.length();
    
    
    // We read char by char. If the read character is }, then we go to the next row. If we read a digit, then we store it to the matrix.
    // Since s[0..1]={{ then we start from s[2].
    for (int k=2;k<length;++k)
    {
       c=s[k];
       switch (c)
       {
	   case '}': 
	      i++;
	      j=1;
	      break;
	   case '0':       
	      A[i][j]=0;
	      j++;
	      break;
	   case '1':
	      A[i][j]=1;
 	      j++;
	      break;
       }
    }
    PathList = new list<Path>;
    PathList3 = new list<list<Path> >;
 
  // Finally we intialize PathSequence[1]=size and the rest we set to zero.
  PathSequence[1]=size;
  for (i=2;i<=size;i++)
    PathSequence[i]=0;
}

// Destructor -- freeing the lists of paths
Graph::~Graph()
{
    delete PathList;
    delete PathList3;
}

// displaying the adjacency matrix
void Graph::showMatrix()
{
   int i,j;
   for (i=1;i<=graphSize;i++)
   {
     for (j=1;j<=graphSize;j++)
        cout<<A[i][j]<<" ";
     cout<<"\n";
   }
}


  
// Generation of all paths....
// algorithm: for any k-element sequence check if it is a path in G. If so, add it to the list


// checking whether a sequence of vertices B[1],...,B[pathSize] is a path in G.
bool Graph::isPath(int B[])
{
    bool wynik=true;
    int i=1;
    int u,v;
    do
    {
        u=B[i],v=B[i+1];
	    wynik=wynik && (A[u][v]==1);
	    i++;
    } while (wynik &&  (i<pathSize));
    
    return wynik;
}


// ***************** Functions for generating permutations *************************************
// The algorithm is taken from the book W. Lipski, "Kombinatoryka dla programistw", WNT, 2004.

//helper: swapping two elements in the table C
void Graph::swap (int i, int j)
{
  int tmp;
  tmp=C[i]; C[i]=C[j]; C[j]=tmp;
}

// Function to reverse the table C[1,...m]
void Graph::reverse(int m)
{
  int i,j;

  i=1;j=m;
  while (i<j)
  {
    swap(i,j);
    ++i;--j;
   }
}
  
//
void Graph::perm (int m)
{
  Path *Q;
  int i;
 
  if (m==1)
  {
    
    // We are in a actual generated permutation of C. If it is a k-path, we add it to the list
    if (isPath(C))
    {
      Q=new Path(C,pathSize);
      PathList->push_back(*Q);
      delete Q;
    }
  }
  
  else
    for (i=1;i<=m;i++)
    {
      perm(m-1);
      if (i<m)
      {
        swap(i,m); reverse(m-1);
      }
      
    }
}

// This function:
// 1. Generates list PathList on which all the k-paths are stored (with repetitions!). At first the old list is removed.
// 2. Sets pathSize=k
void Graph::genPaths (int k)
{
  int i,p;
  int P[k+1];
   
 
   // We remove the old list
  if (PathList!=NULL)
    delete PathList;
  PathList=new list<Path>;
   
  // We initialize pathSize due to the function perm
  pathSize=k;

  // The first sequence
  for (i=1;i<=pathSize;i++)
    P[i]=i;
  
  
  if (k==graphSize)
  {
    for (int t=1;t<=pathSize;t++)
      C[t]=P[t];
    perm(graphSize);
    return ;
  }
  
  p=pathSize;
  
  while (p>=1)
  {
   // We apply all permutations on a given sequence and check, which of the are pathSize paths.
   // Firsty we copy P to C
    for (int t=1;t<=pathSize;t++)
      C[t]=P[t];
   
    perm(pathSize);  
  
   
    if (P[pathSize]==graphSize)
      --p;
    else
      p=pathSize;
    
    if (p>=1)
      for (i=pathSize;i>=p;i--)
	    P[i]=P[p]+i-p+1;
  }
  
}


// checks if the set described by the table C is a k-path vertex cover of G.
bool Graph::isPathVertexCover (int setSize)
{
  bool wynik=true; 

  // For any path we check whether it has a vertex common with the set C
  for (list<Path>::iterator i=PathList->begin(); i!=PathList->end() && wynik; i++)
  {
    wynik=wynik && (i->hasCommonVertex(C, setSize));
  }
 
  return wynik;
}




// Function calculatePsi (int k) calculates psi_k(G) for a given graph. The list of all k-paths MUST BE GENERATED BEFORE using it!
// The calculated value is stored in PathSequence[k].
// Algorithm: We start with s=1 and generate all s-subsets. For any such subset we chceck whether every k-path has a vertex in common with it.
// If so, then psi_k(G)=s. If not, then we continue to generate the subsets.
// WE UTILIZE HERE THE FACT THAT THE SEQUENCE OF PATH NUMBERS IN NON-INCREASING. THEREFORE IT IS NECESSERY TO CALCULATE THE NUMBERS PSI_K
// IN THE FOLLOWING ORDER: PSI_1,PSI_2,...

void Graph::calculatePsi (int k)
{
  int setSize;                          // a candidate for psi_k
  int i,p;
  bool foundSolution=false;
  
  // We utilize the fact that the sequence psi_1(G),...,psi_n(G) is non-increasing

  // We first check whether there are any k-paths. If not, then we set psi_k(G)=0.
  if (PathList->empty())
    PathSequence[k]=0;
  else
  {
    
  
    for (setSize=1; (setSize<=PathSequence[k-1]) && !foundSolution; setSize++)
	{
	    // generate all subsets of a Graph with setSize elements. Actually considered subset is stored in "global" table C.
    
	    for (i=1;i<=setSize;i++)
	      C[i]=i;
  
	    p=setSize;
  
	    while (p>=1)
	    {
	      //here we check if every k-path in Graph has common vertex with C
	      if (isPathVertexCover(setSize))  //if yes, actual subset is a path vertex cover
	      {
	        PathSequence[k]=setSize;
	        foundSolution=true;
	        break;
	      }
          //if not - generate another subset
	      if (C[setSize]==graphSize)
		    p--;
	      else
		    p=setSize;
    
	      if (p>=1)
	        for (i=setSize;i>=p;i--)
	          C[i]=C[p]+i-p+1;
	    }
    
	 }
    }
}

//calculate the values of psi_2(G),...,psi_n(G)
void Graph::calculatePathSequence()
{
  for (int k=2;k<=graphSize;k++)
  {
    genPaths(k);                        //generate paths of k vertices
    calculatePsi(k);                    //calculate psi_k
  }
}

void Graph::showPathSequence()
{
  for (int i=1;i<=graphSize-1;i++)
    cout<<PathSequence[i]<<",";
  cout<<PathSequence[graphSize]<<"\n";
}



//*** VERSION 2

//like in VERSION 1 but vertices are indexed from 0, i.e. P[0],...,P[k-1]
bool Graph::isPath2(int *P, int k)
{
    bool wynik;
    int i=0;

    while(i<k-1)
    {
        if(A[P[i]][P[i+1]]==0)
            return false;
        i++;
    }
    return true;
}

//generate k-paths from the permutation array
void Graph::genPaths2(int k, int** permArray)
{
    Path *Q;	//currently found path

    if(PathList!=NULL)
      delete PathList;
    PathList = new list<Path>;

//number of permutations of a k element set
    long M=fact(k);

//number of k element subsets of the set of vertices
    long N=binom(graphSize,k);    
    for(long i=0; i<N; i++){           //for every subset
	  for(long j=0; j<M; j++)          //and for every sequence of its elements
        if(isPath2(permArray[i*M+j],k)){       //if this sequence forms a path in G - add it to the paths list (note constructor parameters!)
          Q = new Path(permArray[i*M+j], -k);
          PathList->push_back(*Q);
//          Q->showPath();
          delete Q;
          break;                      //leave the "for j" loop - other possible paths generated from this subset have the same set of vertices
                                      // so to compute path vertex covers it is sufficient to add Q to the paths list
      }
   }
   
}

//check if a path (in a graph on nVerts vertices) has a common vertex with the setSize-element set S 
//algorithm: set 1 in V[j] if j is a vertex of P (otherwise 0), check if V[p]==1 for any element p of S
// complexity: |V|+|P|+|S| <= 2*|V| where |V| is the number of vertices in a graph
inline bool Path::hasCommonVertex2(int *S, int setSize, int nVerts)
{
  int V[nVerts+1];
  for(int i=0; i<=nVerts; i++)
    V[i]=0;
  for(int j=1; j<=size; j++)
    V[P[j]]=1;
  for(int j=0; j<setSize; j++)
    if(V[S[j]]==1)
      return true;
  return false;
}

//like in VERSION 1
bool Graph::isPathVertexCover2(int *S, int setSize)
{
    bool wynik; 
    for (list<Path>::iterator i=PathList->begin(); i!=PathList->end(); i++) //for every path on the list
        if(!i->hasCommonVertex2(S, setSize, graphSize))     //check if it has a common vertex with S
            return false;
    return true;
}

//calculating the value of psi_k - check every subset to find the minimum k-path vertex cover
int Graph::calculatePsi2(int k, int*** setArray)
{
//if there are no paths of length k - return 0
    if(PathList->empty())
        return 0;

//there is a path of length k, so check subsets of size 1..psi_{k-1}(G) (the sequence psi is nonincreasing)
    for(int setSize=1; setSize<=PathSequence[k-1]; setSize++)
    {
    //number of setSize element subsets of the set of vertices
        long N=binom(graphSize,setSize);
        for(long i=0; i<N; i++)      //check all subsets...
            if(isPathVertexCover2(setArray[setSize][i], setSize)) // ...given in setArray[setSize] - if i'th is a vertex cover...
                return setSize;    // ...then return its size
    }
//	cout<<"-1"<<endl;
    return -1;        //should never happen
}

//calculate the path sequence
void Graph::calculatePathSequence2(int*** setArray, int*** permArray)
{
    for(int k=2; k<=graphSize; k++){                //for increasing values of k
        genPaths2(k, permArray[k]);                 // generate all paths on k vertices in G (from the table of permutations)
        PathSequence[k]= calculatePsi2(k, setArray);// and calculate psi_k
    }
}


//VERSION 3

//return first vertex of a path
int Path::firstVertex(){
   return P[1];
}
//return last vertex of a path
int Path::lastVertex(){
   return P[size];
}

//add vertex v at the beginning of a path
void Path::addFirstVertex(int v){
  for(int i=1,j=size; i<j; i++,j--)  // reverse the path
  {
      int tmp = P[i]; P[i]=P[j]; P[j]=tmp;
  }
  addLastVertex(v);     //add v at the end of the reversed path
}
//add vertex v at the end of a path
void Path::addLastVertex(int v){
  P[++size] = v;
  hash|= 1<<(v-1);
}

//check if a path contains vertex v (using hash value)
bool Path::contains(int v){
    return (hash&(1<<(v-1)));
}

//add new path to the list of lists
void addPath3(list<list<Path> > *list3, Path p)
{
    for(list<list<Path> >::iterator i3=list3->begin(); i3!=list3->end(); i3++)
        if((i3)->front().getHash()==(p).getHash())  //if hash value of p equals to some hash value on the list
        {
            (i3)->push_front(p);                    // insert p to the corresponding list
            return;                                 //  and exit
        }
    list<Path> newList;                             //if not - create new list of paths
    list3->push_front(newList);                     // insert it to the list of lists
    list3->front().push_front(p);                   // and insert p to the newly created list
}

void Graph::genPaths3(int k)
{
  int i,j;
  list<list<Path> > *newList3;
  int vert[2];
  
//we do not calculate the case of 1-vertex cover
  if (k<=1)
    return;

  newList3 = new list<list<Path> >;     //create empty list of lists

  if (k==2)                             //k=2: (standard) vertex cover
    for(i=1; i<=graphSize; i++)
      for(j=i+1; j<=graphSize; j++)
        if(A[i][j]){                    //if there is an edge i-j
          vert[0]=i; vert[1]=j;         //(vertices in edges are ordered: i<j)
          Path p(vert, -2);             // create a path on 2 vertices from the edge
          addPath3(newList3, p);        // and insert it to the list
        }

  if (k>2)                              //k>2: k-path vertex cover
  for (list<list<Path> >::iterator i3=PathList3->begin(); i3!=PathList3->end(); i3++)   //for any list on the list of lists
    for (list<Path>::iterator it=(i3)->begin(); it!=(i3)->end(); it++)                  // and for any path on this list
      for(i=1; i<=graphSize; i++)
        for(j=1; j<=graphSize; j++)
        {
          //if there is an edge i-j AND path ends in vertex i AND path does not contain j
          // AND adding j after i preserves the order of vertices (first vertex < last vertex)
          if(A[i][j] && (it->lastVertex()==i) && !it->contains(j) && (it->firstVertex()<j))
          {
            Path p = Path(*it);         //create a copy of current path
            p.addLastVertex(j);         // add j as the last vertex
            addPath3(newList3, p);      // insert new path to the list
            continue;                   // continue loop
          }
          //if there is an edge i-j AND path begins in vertex i AND path does not contain j
          // AND adding j before i preserves the order of vertices (first vertex < last vertex after reversing the path)
          if(A[i][j] && (it->firstVertex()==i) && !it->contains(j) && (j>it->lastVertex()))
          {
            Path p = Path(*it);         //create a copy of current path
            p.addFirstVertex(j);        // add j as the first vertex and reverse the path
            addPath3(newList3, p);      // insert new path to the list
          }
        }

  delete PathList3;                     //delete list of lists for the previous value of k
  PathList3 = newList3;                 // store list of lists calculated for the current value of k to PathList3
/*
  cout<<endl<<"Paths of length "<<k<<": "<<endl;
  for (list<list<Path> >::iterator i3=PathList3->begin(); i3!=PathList3->end(); i3++)
  {
      for (list<Path>::iterator i=(i3)->begin(); i!=(i3)->end(); i++)
      {
        i->showPath();
        cout<<"  ";
      }
      cout<<"*"<<endl;
  }
*/
}

//check if a path has a common vertex with the set on setSize vertices (its hash value is s)
//algorithm: bitwise and of hash values
// complexity: constant (or at most |P|+|S|<=|V| if both hash values are calculated "from scratch" before EVERY call to this function)
bool Path::hasCommonVertex3(hash_t s, int setSize)
{
    return (hash&s);
}

//return the value of hash function for a path
hash_t Path::getHash()
{
    return hash;
}

//like in VERSION 1
bool Graph::isPathVertexCover3(hash_t set, int setSize)
{
    for (list<list<Path> >::iterator i3=PathList3->begin(); i3!=PathList3->end(); i3++) //for every list on the list of lists
        if(!(i3)->front().hasCommonVertex3(set, setSize))   //check if the first path on the list has a common vertex with the given set
                                                            // we only need to check one path since all other paths on the list have the same hash value = the same set of vertices
            return false;
    return true;
}


//calculating the value of psi_k - check every subset to find the minimum k-path vertex cover
int Graph::calculatePsi3(int k)
{
//if there are no paths of length k - return 0
    if(PathList3->empty())
        return 0;

//there exists a path of length k, so check subsets of size 1..psi_{k-1}(G) (the sequence psi is nonincreasing)
    for(int setSize=1; setSize<=PathSequence[k-1]; setSize++)
    {
    //number of setSize-element subsets of the set of vertices
        long N=binom(graphSize,setSize);
        hash_t set = firstSet(graphSize,setSize);   //hash value of the first set to consider
        for(long i=0; i<N; i++)                     //check all subsets:
        {
            if(isPathVertexCover3(set, setSize))    // if current is a k-path vertex cover
                return setSize;                     //  then return its size
            set = nextSet(set, graphSize, setSize); // if not, generate hash value of the next set
        }
    }
//	cout<<"-1"<<endl;
    return -1;        //should never happen
}

//calculate the values of psi_2(G),...,psi_n(G)
void Graph::calculatePathSequence3()
{
    for(int k=2; k<=graphSize; k++){        //for increasing values of k
        genPaths3(k);                       // generate all paths on k vertices in G (VERSION 3, list of lists)
        PathSequence[k]=calculatePsi3(k);   // and calculate psi_k
    }
}

