#include "komb.h"
using namespace std;

//value of the factorial function
long fact(int n)
{
    return (n<=0) ? 1L : n*fact(n-1);
}

//value of the binomial coefficient
long binom(int n, int k)
{
    if(k==0) return 1;
    if(n==0) return 0;
    return n*binom(n-1,k-1)/k;
}

//insert element x on position pos into array A and save the result (larger array) in B
void insert(int x, int pos, int *A, int n, int *B){
    int i;
    for(i=n; i>pos; i--)
        B[i]=A[i-1];
    B[pos]=x;
    for(i=pos-1; i>=0; i--)
        B[i]=A[i];
}



//saves all k-element combinations of the set 1..n in array w
void genComb(int n, int k, int** w)
{
    int set[n];
    for(int i=0; i<n; i++)
        set[i]=i+1;

    combine(set, n, k, w);
}

//saves all k-element combinations of the n-element array S in array w
void combine(const int *S, int n, int k, int** w)
{
//what if n<k?
    if(n<k)
        return;
        
//1-element combinations - there are n of them; save elements of S to proper positions in w
    if(k==1){
        for(int i=0; i<n; i++)
            w[i][0]=S[i];
        return;
    }
    
//n-element combination - there is just one; save the whole set S to proper position
    if(k==n){
        for(int i=0; i<n; i++)
            w[0][i]=S[i];
        return;
    }

//the number of (k-1)-element combinations of (n-1) elements =
// = the number of k-element combinations of S that contain the element S[n-1]
    long B = binom(n-1, k-1);
//generate (k-1)-element combinations of S[0],..., S[n-2]
    combine(S, n-1, k-1, w);
    for(long i=0; i<B; i++)                        //for each generated combination
        insert(S[n-1], k-1, w[i], k-1, w[i]);	  // insert the element S[n-1] onto the last position
//generate k-element combinations of S[0],..., S[n-2] ( = combinations of S not containing the element S[n-1])
    combine(S, n-1, k, w+B);
}



//saves all permutations of the set 1..n in array w
void genPerm(int n, int** w)
{
    int set[n];
    for(int i=0; i<n; i++)
        set[i]=i+1;

    permute(set, n, w);
}

//saves all permutations of the n-element array S in array w
void permute(const int* S, int n, int** w)
{
//what if n<=0?
	if(n<=0)
		return;

//if the set contains a single element, there is only 1 permutation - write it to w[0]
    if(n==1){
        w[0][0]=S[0];
        return;
    }

//number of permutations to generate-1
    long curr = fact(n)-1;
//generate permutations of S[0]..S[n-2]
    permute(S, n-1, w);
    for(long per=fact(n-1)-1; per>=0; per--)				  //then, for all generated permutations
        for(int pos=0; pos<n; pos++, curr--)			  // and all possible positions 0..n-1
            insert(S[n-1], pos, w[per], n-1, w[curr]);    // insert the element S[n-1] onto this position in permutation
}


// hash value of the "first" k-element subset of 1..n =
// = in binary: k 1's, followed by (n-k) 0's
hash_t firstSet(int n, int k)
{
    hash_t s = 0;
    hash_t m = 1<<(k-1);
    while(m){
        s|= m;
        m>>=1;
    }
    return s<<(n-k);
}

// hash value of the k-element subest of 1..n following the subset represented by s
hash_t nextSet(hash_t s, int n, int k)
{
    int l=n;            // currently considered element of the set 1..n = bit in the hash value
    hash_t m = 1;       //  and its associated hash value (single bit set)
    if(s&m){            //if s contains the last element, for example s={1,2,5,6} for n=6 and k=4
        while(s&m){     // find the element not in s (decrement l and modify m), for ex. l=4
            m<<=1;
            l--;
        }
        if(n-l==k)      //if there are 1's on k lowest bits of s
            return 0;   // all sets were generated, exit
        while(!(s&m))   //if not, find the first element strictly smaller than l that is in s, for ex. 2
            m<<=1;
        s^=m;           //  remove it from s (set its bit to 0)
        m>>=1;
        while(l<=n){    //  insert next (n-l+1) elements into the set, hash s is now for ex. {1,3,4,5,6}
            s|=m;
            m>>=1;
            l++;
        }
        while(m){       // remove remaining elements from the set, calculated hash for the next set is for ex. {1,3,4,5}
            if(s&m) s^=m;
            m>>=1;
        }
    }else{              //if s does not contain the last element, for example s={1,2,4,5} for n=6 and k=4
        while(!(s&m))   // find the largest element in s, for ex. 5
            m<<=1;
        s^= m;          //remove it from the set
        s|= (m>>1);     // and insert next value, current set is for ex. {1,2,4,6}
    }
    return s;
}

